# -*- coding: utf-8 -*-


def test_example_controller():

    from kiara.examples.example_controller import (
        execute_pipeline_with_example_controller,
    )

    execute_pipeline_with_example_controller()
