# %% [markdown]
# # Table of contents
# * [Assembling the workflow](#assembly)
#     * [Creating the steps of the workflow](#creating_steps)
{%  for stage in structure.processing_stages -%}
#         * [Execution stage: {{ loop.index }}](#execution_stage_{{ loop.index }})
{% endfor -%}
#     * [Connecting the workflow steps](#connecting_steps)
# * [Workflow information](#workflow_info)
# * [Setting workflow inputs](#workflow_inputs)
# * [Workflow results](#workflow_results)
{% for stage in structure.processing_stages -%}
{% for step_id in stage -%}
{%  set step = structure.get_step(step_id) -%}
#     - [Step: ``{{ step.step_id }}``](#output_stage_{{ loop.index }}_{{ step.step_id }})
{% endfor -%}
{% endfor -%}
